class Explosion extends GameObject {
  Vec2 position;
  float radius;
  float impulse;
  float damage;
  float impulseFalloff;
  float damageFalloff;
  float lifetime;
  float remainingLifetime;
  color fillColor = color(255, 100, 100);
  
  GameObject creator = null; //What/who caused this explosion. Can be null.
  
  protected boolean damageDone = false;
  
  Explosion(){
    radius = 1;
    impulse = 4000;
    damage = 5;
    impulseFalloff = 1;
    damageFalloff = 1;
    remainingLifetime = lifetime = 5;
  }
  
  void setLifetime(float lifetime){
    this.remainingLifetime = this.lifetime = lifetime;
  }
  
  void update(float dt){
    if ( lifetime > 0 ){
      remainingLifetime -= dt;
    }
    
    if ( remainingLifetime <= 0 ){
      die();
    }
    
	if ( !damageDone ){
		WorldQuery targets = queryEntitiesInRange();
		for(PhysicalObject target : targets){
		  if (target.dead || (target.body == null) || !(target instanceof Ship) ){
			continue;
		  }
		  
		  Ship targetShip = (Ship)target;
		  
		  Vec2 direction = targetShip.getPosition().sub(position);
		  float distance = direction.length();
		  
		  //Check whether the target is *really* in range
		  if ( distance > radius + targetShip.radius ){
			continue;
		  }
		  		  
		  direction.normalize();
		  targetShip.body.applyImpulse(direction.mul(impulseAtDistance(distance)), targetShip.getPosition());
		  
		  float scaledDamage = damageAtDistance(distance);
		  targetShip.takeDamage(scaledDamage, creator, EXPLOSIVE_DAMAGE);
		}
		damageDone = true;
	}
  }
  
  void draw(){
    pushMatrix();
    translate(position.x, position.y);
    float a = map(max(remainingLifetime / lifetime, 0), 0, 1, 5, 200);
    fill(fillColor, a);
    noStroke();
    ellipse(0, 0, radius * 2, radius * 2);
    popMatrix();
  }
  
  WorldQuery queryEntitiesInRange(){
    AABB bounds = new AABB(new Vec2(position.x - radius, position.y - radius), new Vec2(position.x + radius, position.y + radius));
    return new AABBQuery(bounds, world, false, 50);
  }
  
  float damageAtDistance(float distance){
    return damage * ( distance / radius * damageFalloff );
  }
  
  float impulseAtDistance(float distance){
    return impulse;
  }
}
