class GameObject {
  boolean dead = false;
  float timeOfDeath = 0; //timestamp, seconds
  float stickAroundTime = 0; //How long to stick around after death
  float remainingStickAroundTime = 0;
  
  boolean drawOnUI = false; //Whether to call the object's drawUI() method when rendering the UI. Only set this flag once!
                            //The director doesn't re-check it after the entity has been added.
  
  void handleInput(){
  }
  
  void update(float dt){
  }
  
  void draw(){
  }
  
  void drawUI(){
  }
  
  void deInit(){
    //stub
  }
  
  void die(){
    dead = true;
    timeOfDeath = millis() / 1000;
    remainingStickAroundTime = stickAroundTime;
  }
}

class EntityBag {
  private List<GameObject> entities, entitiesWithUI, entitiesToAdd, entitiesToRemove;
 
  EntityBag(){
    entities = new ArrayList<GameObject>();
	entitiesWithUI = new ArrayList<GameObject>();
    entitiesToAdd = new ArrayList<GameObject>();
    entitiesToRemove = new ArrayList<GameObject>();
  }
  
  void add(GameObject entity){
    entitiesToAdd.add(entity);
	if ( entity.drawOnUI ){
		entitiesWithUI.add(entity);
	}
  }
  
  void remove(GameObject entity){
    entitiesToRemove.add(entity);
  }
  
  List<GameObject> getAll(){
    return entities;
  }
  
  void update(float dt){
    float currentTime = millis() / 1000.0;
    
    for(GameObject entity : entities){
      entity.update(dt);
      if ( entity.dead ){
        entity.remainingStickAroundTime -= dt;
        if ( entity.remainingStickAroundTime <= 0 ){
          entitiesToRemove.add(entity);
        }
      }
    }
    
    //Warning: Inefficient! Consider optimization later.
    for(GameObject entity : entitiesToRemove){
	  entity.deInit();
      entities.remove(entity);
	  if ( entity.drawOnUI ){
		entitiesWithUI.remove(entity);
	  }
    }
    entitiesToRemove.clear();
    
    for(GameObject entity : entitiesToAdd){
	  if ( entity instanceof PhysicalObject ){
		((PhysicalObject)entity).addTo(world);
	  }
      entities.add(entity);
    }
    entitiesToAdd.clear();
  }
  
  void draw(){
    for(GameObject entity : entities){
      entity.draw();
    }
  }
  
  void drawUI(){
    for(GameObject entity : entitiesWithUI){
      entity.drawUI();
    }
  }
  
  void handleInput(){
    for(GameObject entity : entities){
      entity.handleInput();
    }
  }
}
