final static int UNKNOWN_DAMAGE = 0;
final static int EXPLOSIVE_DAMAGE = 1;
final static int COLLISION_DAMAGE = 2;
final static int KAMIKAZE_DAMAGE = 3;

class Ship extends PhysicalObject {
  float radius;
  float maxThrust = 10000;
  float maxSpeed = 100;
  boolean isTractored = false;

  float hitpoints = 0;
  float maxHitpoints = 100;
  
  float explosionDamage = 40;
  float explosionRadiusFactor = 1.7;
  float explosionImpulse = 5000;
  
  color baseColor;
  
  Ship(){
    //stub
    baseColor = color(255, 255, 0);
  }
  
  void takeDamage(float amount, GameObject damageDealer, int damageType){
    if ( dead ) return;
    
    hitpoints = hitpoints - amount;
    if ( hitpoints <= 0 ){
      die(amount, damageDealer, damageType);
    }
  }
  
  void takeDamage(float amount){
    takeDamage(amount, null, UNKNOWN_DAMAGE);
  }
  
  void die(float fatalDamage, GameObject damageDealer, int damageType){
	hitpoints = 0;
	super.die();
  }
  
  void die(){
    die(0, null, UNKNOWN_DAMAGE);
  }
  
  void explode(float multiplier){
	Explosion boom = new Explosion();
    boom.radius = constrain(radius * explosionRadiusFactor * multiplier, 1, 30);
    boom.damage = constrain(explosionDamage * sqrt(multiplier), 1, 500);
    boom.impulse = explosionImpulse;
    boom.position = getPosition();
    boom.setLifetime(0.5);
	boom.creator = this;
    entities.add(boom);
  }
  
  void onCollisionResult(ContactResult result, PhysicalObject otherObject){
    if (dead){
      return;
    }
    
    //Ignore collisions with other dead things
    if ( otherObject != null && otherObject.dead ){
      return;
    }
    
    //Maybe take damage from the collision
    float damage = getCollisionDamage(result, otherObject);
          
    if ( damage > 1 ){
      takeDamage(damage, otherObject, COLLISION_DAMAGE);
    }
  }
  
  float getCollisionDamage(ContactResult result, PhysicalObject otherObject){
    if ( body == null ){
      return 0;
    }
    
    //Take no damage from the player's ship if currently tractored by it
    if ( isTractored && (otherObject instanceof PlayerShip) ){
      return 0;
    }
    
    float damage = result.normalImpulse / 8000 * (1 - getArmor());
    //Take less collision damage from scenery, i.e. walls.
    if ( otherObject == null ){
      damage = damage * 0.5;
    }
    
    return damage;
    /*
    float energy = abs(collisionEnergy(result));
    float damage = energy / 1000000;
    return damage;
    */
  }
  
  float getArmor(){
    return 0.0;
  }
  
}


