import processing.core.*; 
import processing.xml.*; 

import org.jbox2d.dynamics.contacts.*; 
import org.jbox2d.collision.*; 
import org.jbox2d.common.*; 
import org.jbox2d.dynamics.joints.*; 
import org.jbox2d.dynamics.*; 
import java.awt.Robot; 
import java.awt.AWTException; 
import java.awt.event.InputEvent; 
import java.awt.MouseInfo; 
import java.awt.Point; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class SlingPrototype extends PApplet {









 



int gameState = 0; //0 = intro, 1 = in game, 2 = game over
boolean paused = false;

World world;
Vec2 gravity = new Vec2(0.0f, 0.0f);
final float worldToScreenScale = 7;
float worldHalfWidth, worldHalfHeight;

Body[] walls;
final float wallThickness = 1.0f;

PlayerShip playerShip;
final float shipRadius = 2.5f; //meters
final float shipDensity = 450; //kg/m^2
final float mouseStrength = 8.2f * (PI * shipRadius * shipRadius * shipDensity);
final float playerHealth = 500;

final float hookRadius = 1.0f;
final float hookDensity = shipDensity / 3;

final float initialBeamLength = hookRadius * 3.5f;
final float maxBeamLength = 15.0f;
final float minBeamLength = 0.5f;
final float beamExtensionSpeed = 0.4f;
final float beamContractionSpeed = 0.4f;

final float repulsorCost = 200;

final float enemyShipDensity = 100;

float score = 0;
int enemiesStillAlive = 0;

//Entities
EntityBag entities;
Director director;

//Physics
List<BotContact> possibleTargets;
MyContactListener contactHandler;

//Mouse control
Robot robot;
Point mouse, prevMouse;
float mouseDx, mouseDy;

PFont font, scoreFont;
ParticleEmitter collisionEmitter, textEmitter;

float gameTime = 0; //seconds
long lastTickTime = 0; //nanoseconds

public void setup(){
  size(1024, 800);
  smooth();
  frameRate(60);
  fill(255);
  
  worldHalfWidth = width / 2 / worldToScreenScale;
  worldHalfHeight = height / 2 / worldToScreenScale;
  
  //Fonts
  font = loadFont("AgencyFB-Reg-48.vlw");
  textFont(font, 20);
  scoreFont = loadFont("AgencyFB-Bold-36.vlw");
  
  //Mouse
  try {
    robot = new Robot();
  } 
  catch (Exception e) {
    println("Can't Initialize the Robot");
  }
  addMouseWheelListener(new java.awt.event.MouseWheelListener() { 
    public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) { 
      mouseWheel(evt.getWheelRotation());
  }}); 
  
}

public void draw(){
  switch(gameState){
    case 0: intro(); break;
	case 1: runGame(); break;
	case 2: gameOver(); break;
  }
}

public void newGame(){
	gameState = 1;
	noCursor();
	prevMouse = MouseInfo.getPointerInfo().getLocation(); ;

	possibleTargets = new ArrayList<BotContact>();
	initPhysics();

	//Create entities
	entities = new EntityBag();

	director = new Director();
	entities.add(director);
	
	collisionEmitter = new ParticleEmitter(new CollisionParticleFactory());
	collisionEmitter.pVelocity = new FloatRange(40, 50);
	collisionEmitter.pLifetime = new FloatRange(0.2f, 0.6f);
	collisionEmitter.pDamping = new FloatRange(0.05f, 0.3f);
	collisionEmitter.rate = 0;
	entities.add(collisionEmitter);
	
	textEmitter = new ParticleEmitter(new TextParticleFactory());
	textEmitter.pVelocity = new FloatRange(1, 3);
	textEmitter.pLifetime = new FloatRange(2, 5);
	textEmitter.pDamping = new FloatRange(0.05f, 0.1f);
	textEmitter.pDirection = new VectorRange(new Vec2(0, -1), 0);
	textEmitter.rate = 0;
	entities.add(textEmitter);
	
	playerShip = new PlayerShip();
	entities.add(playerShip);
	entities.add( new HealthBar(playerShip) );
	entities.add( new PowerBar(playerShip) );
	
	paused = false;
	gameTime = 0;
	lastTickTime = System.nanoTime();
}

public void runGame(){
  final long thisTickTime = System.nanoTime();
  final float dt = (thisTickTime - lastTickTime) / 1000000000f;
  if ( dt <= 0 ){
	return;
  }
  lastTickTime = thisTickTime;
  
  if ( paused ){
	return;
  }
  gameTime += dt;
  
  handleInput();
  entities.update(dt);
  updatePhysics(dt);
  
  background(255);
  drawWorld();
  drawUI();
}

public void goToGameOver(){
	cursor();
	gameState = 2;
}

public void gameOver(){
	background(255);
	fill(0);
	noStroke();
	textAlign(CENTER, BASELINE);
	
	pushMatrix();
	translate(width/2, height/2 - 50);
	
	textSize(50);
	text("GAME OVER", 0, 0);
	textSize(24);
	text("SCORE: " + ceil(score), 0, 32);
	popMatrix();
	
	fill(100);
	text("Press SPACE to start a new game or ESC to exit", width/2, height-20);
}

public void goToIntro(){
	cursor();
	gameState = 0;
}

public void intro(){
	background(255);
	fill(0);
	noStroke();
	textAlign(CENTER, BASELINE);
	
	pushMatrix();
		translate(width/2, height/2 - 50);
		textSize(50);
		text("Sling Prototype", 0, 0);
	popMatrix();
	
	fill(100);
	textSize(24);
	text("Press SPACE to start the game", width/2, height-20);
}

public void keyReleased(){
	if ( key != ' ' ){
		return;
	}

	switch(gameState){
		case 0: //Intro => Play
			newGame();
			break;
		case 2: //Game Over => Intro
			goToIntro();
			break;
	}
}

public void keyPressed(){
	if ( gameState == 1 ){
		if ( (key == 'p') || (key == 'P') ){
			paused = !paused;
			if ( paused ){
				cursor();
			} else {
				noCursor();
			}
		}
	}
}

public void awardPoints(int amount, int comboLevel, Vec2 position){
	score += amount;
	
	if ( (playerShip != null) && !playerShip.dead ){
		playerShip.gainPower(amount / 10);
	}
	
	if ( (position != null) && (amount > 0) ){
		TextParticle particle = (TextParticle)(textEmitter.emitParticle());
		particle.content = "+" + Integer.toString(amount);
		particle.size = constrain(ceil(map(sqrt(amount), 3, 15, 18, 48)), 18, 64);
		particle.lifetime = map(sqrt(amount), 2, 15, 1, 7);
		particle.position.set(position);
	}
}
class Bot extends Ship {
  Weapon weapon;
  
  protected float density = enemyShipDensity; //kg / m^2
  protected float friction = 0.3f;
  protected float restitution = 0.50f;
  protected final float epsilon = 0.00001f;
  
  Vec2 steeringVector;
  boolean canPursue = false;
  int points = 1;
  
  boolean aboutToExplode = false;
  float explosionCountdown = 0;
  int explosionComboCount = 1;	//Used for combo tracking. Dead bots set this when they blow up, 
                                //and living ones check it when they're hit by that explosion.

  Bot(){
    this(new Vec2(0, 0), 1, 2);
  }
  
  Bot(Vec2 startPosition, float radius, float maxHitpoints){
    this.startPosition = startPosition.clone();
    this.radius = radius;
    this.maxHitpoints = maxHitpoints;
    this.hitpoints = maxHitpoints;
    
    steeringVector = new Vec2(0, 0);

    stickAroundTime = 5;
    maxThrust = 90000;
    maxSpeed = 40;
	
	enemiesStillAlive++;
  }
  
  Bot(Vec2 startPosition, float radius){
    this(
      startPosition, radius, 
      ceil(map(radius, 1, 4, 50, 200)/10)*10
    );
  }
  
  public void addTo(World world){
    BodyDef bodydef = new BodyDef();
    bodydef.position.set(startPosition);
    startPosition = null; //we won't need it any more
    
    bodydef.angularDamping = 0.2f;
    //bodydef.linearDamping = 0.1;
  
    CircleDef shapedef = new CircleDef();
    shapedef.radius = radius;
    shapedef.density = density;
    shapedef.friction = friction;
    shapedef.restitution = restitution;
  
    body = world.createBody(bodydef);
    body.createShape(shapedef);
    body.setMassFromShapes();
    body.setUserData(this);
  }
  
  public void update(float dt){
    if ( dead ){
      return;
    }
	
	if ( aboutToExplode && !isTractored ){
		explosionCountdown -= dt;
		if ( explosionCountdown <= 0){
			die();
		}
	}
    
    if ( !isTractored ){
		steeringVector = calculateSteeringVector();
		//Limit the steering force to our maximum thrust
		if ( steeringVector.length() > maxThrust ){
			steeringVector.normalize();
			steeringVector.mulLocal(maxThrust);
		}
		body.applyForce(steeringVector, body.getWorldCenter());
      
		//Close enough to shoot?
		if ( weapon != null ){
			float targetDistance = playerShip.getPosition().sub(getPosition()).length();
			if ( targetDistance < getEngagementDistance() ) {
				weapon.fire();
			}
		}
    }
    
    if ( weapon != null ){
      weapon.update(dt);
    }
  }
  
  public Vec2 calculateSteeringVector(){
	Vec2 steering = new Vec2(0, 0);
	
	if ( canPursue ){
		//"Intelligently" intercept the player
		steering.addLocal( pursue(playerShip) );
	} else {
		//Charge directly at the player
		steering.addLocal( seek(playerShip.getPosition()) );
	}
	
	//Avoid collisions with other bots
	steering.addLocal( separate(6, entities.getAll()).mul(2) );
	  
	return steering;
  }
  
  public Vec2 steer(Vec2 target, float brakingDistance){
    float control = 0.10f;
    float dt = 1.0f / 60.0f; // The (average) physics step.

    Vec2 steer;
    
    Vec2 desiredVelocity = target.sub(getPosition());
    float distance = desiredVelocity.length();
    
    if (distance > 0) {
        // Compute the aiming direction
        desiredVelocity.normalize();
        
        if ( (brakingDistance == 0) || (distance > brakingDistance) ){
          // The ship's ideal velocity is the direction to the target multiplied by the max speed
          desiredVelocity.mulLocal(maxSpeed);
        } else {
          // Same as above, except the ships gradually slows down to a stop as it approaches the target
          desiredVelocity.mulLocal(maxSpeed * distance / brakingDistance);
        }        

        // Get the current ship's velocity because we will apply a force to compensate this.
        Vec2 currentVelocity = body.getLinearVelocity();

        // Compensate the current velocity by the desired velocity, based on the control factor
        Vec2 finalVelocity = desiredVelocity.sub(currentVelocity).mul(control);

        // Transform our velocity into an impulse (get rid of the time and mass factor)
        steer = finalVelocity.mul(body.getMass() / dt);
        
        if ( steer.length() > maxThrust ){
          steer.normalize();
          steer.mulLocal(maxThrust);
        }
    } else {
      steer = new Vec2(0, 0);
    }
    
    return steer;
  }
  
  public Vec2 steer(Vec2 target){
    return steer(target, 0);
  }
  
  public Vec2 seek(Vec2 target){
    return steer(target);
  }
  
  public Vec2 flee(Vec2 target){
    return steer(target).negate();
  }
  
  public Vec2 arrive(Vec2 target, float brakingDistance){
    return steer(target, brakingDistance);
  }
  
  public Vec2 separateOld(float maxDistance, List<GameObject> flock){
    Vec2 accumulatedForce = new Vec2(0, 0); 
    float dt = 1.0f / 60.0f;
    
    for(GameObject thing: flock){
      if (!(thing instanceof Bot)){
        continue;
      }
      Bot bot = (Bot)thing;
      Vec2 separationForce = getPosition().sub(bot.getPosition()); 
      
      //Check if the target is too far to matter
      float distance = separationForce.length() - radius - bot.radius;
      if ( distance > maxDistance ){
        continue;
      }
      //Steer away from the target
      separationForce.normalize();
      //The force is proportional to how close we are to the target
      Float r = max(distance, 0.0001f) / maxDistance;
      separationForce.mulLocal( 1 / (r * r) );
      
      accumulatedForce.addLocal(separationForce);
    }
    
    //return steer(accumulatedForce.addLocal(getPosition()));
    //*
	//accumulatedForce.normalize();
    accumulatedForce.mulLocal(maxSpeed);
    accumulatedForce.mulLocal(body.getMass() / dt);

    //limitVector(accumulatedForce, maxThrust);
    return accumulatedForce;
  }
  
  public Vec2 separate(float desiredSeparation, List<GameObject> flock, Class neighbourType){
	float dt = 1.0f / 60.0f;
    Vec2 steer = new Vec2(0, 0); 
    int count = 0;
    
    for(GameObject thing: flock){
      if (!(neighbourType.isInstance(thing)) || (thing == this)){
        continue;
      }
	  assert (thing instanceof Ship) : "Can't execute separation behaviour against an entity that isn't a Ship";
	  
      Ship bot = (Ship)thing;
	  
      Vec2 diff = getPosition().sub(bot.getPosition()); 
	  float distance = diff.length() - radius - bot.radius;
	  
	  if ( (distance > 0) && (distance < desiredSeparation) ){
		diff.normalize();
		diff.mulLocal( 1f / (distance * distance) ); //or 1 / d^2
		steer.addLocal(diff);
		count++;
	  }
    }
	
	if ( steer.length() > 0 ){
		steer.normalize();
		steer.mulLocal(maxSpeed);
		steer.subLocal(body.getLinearVelocity());
		
        steer.mulLocal(body.getMass() / dt);
        
        if ( steer.length() > maxThrust ){
          steer.normalize();
          steer.mulLocal(maxThrust);
        }
	}
	
	return steer;
  }
  
  public Vec2 separate(float desiredSeparation, List<GameObject> flock){
	return separate(desiredSeparation, flock, Bot.class);
  }
  
  // For every nearby ship in the system, calculate the average velocity
  public Vec2 align (float neighbourDistance, List<GameObject> flock) {
	float dt = 1f / 60f;
    Vec2 steer = new Vec2(0, 0);
    int count = 0;
	
    for (GameObject thing : flock) {
		if ( !(thing instanceof Bot) || (thing == this) ){
			continue;
		}
		
		Bot bot = (Bot)thing;
		float distance = getPosition().sub(bot.getPosition()).length();
		
		if ( distance < neighbourDistance ){
			steer.addLocal(bot.body.getLinearVelocity());
			count++;
		}
    }
	
    if (count > 0) {
		steer.mulLocal(1 / (float)count);
    }

    // As long as the vector is greater than 0
    if (steer.length() > 0) {
		steer.normalize();
		steer.mulLocal(maxSpeed);
		steer.subLocal(body.getLinearVelocity());
		
		steer.mulLocal(body.getMass() / dt);
		
		if ( steer.length() > maxThrust ){
		  steer.normalize();
		  steer.mulLocal(maxThrust);
		}
    }
	
    return steer;
  }
  
  public Vec2 pursue(Ship target){
	float distance = getPosition().sub(target.getPosition()).length();
	float lookAhead = distance / target.maxSpeed;
	Vec2 futureLocation = target.getPosition().add(target.body.getLinearVelocity().mul(lookAhead));
	return seek(futureLocation);
  }
  
  //Draw the bot ship
  public void draw(){
    pushMatrix();
      Vec2 position = body.getPosition();
      translate(position.x, position.y);
      
      if ( !dead ){
        //drawHitpoints();
        drawSteeringVector();
      }

      //Display the body
      rotate(body.getAngle());
      drawHull();
      
      //Display the weapon
      if ( weapon != null && !dead ){
        weapon.draw();
      }
	  
	  //Some sort of visual marker for when we're about to explode
	  if ( !dead && aboutToExplode ){
		noFill();
		stroke(255, 20, 20);
		final float effectRadius = max(radius * 2.4f, radius*2 + 1.5f);
		ellipse(0, 0, effectRadius, effectRadius);
	  }
    popMatrix();
  }
  
  public void drawHull(){
    if ( !dead ){
      stroke(0);
      fill(255);
    } else {
      stroke(180, getDeathlyAlpha());
      noFill();
    }
    
    ellipse(0, 0, radius*2, radius*2);
    rectMode(CENTER);
    rect(0, -radius/2, 1.0f, radius);
    fill(255, 50, 50);
    
    //Display hitpoints as a growing bar
    if ( !dead ){
      noStroke();
      rect(0, -radius/2, 0.8f, map(hitpoints, 0, maxHitpoints, radius, 0) );
    }
  }
  
  public void drawHitpoints(){
    fill(0);
    textAlign(CENTER, BASELINE);
	textFont(font);
    
    pushMatrix();
      translate(0, - radius - 6.0f/worldToScreenScale);
      scale(1.0f/worldToScreenScale); //Necessary to prevent horrible text distortion
      textSize(24);
      text(ceil(hitpoints), 0, 0);
    popMatrix();
  }
  
  public void drawSteeringVector(){
	final float visibleLength = radius + map(steeringVector.length()/maxThrust, 0, 1, 0, 2.5f);
	Vec2 visibleVector = steeringVector.clone();
	visibleVector.normalize();
	visibleVector.mulLocal(visibleLength);
	
    stroke(50, 50, 255);
    line(0, 0, visibleVector.x, visibleVector.y);
  }
  
  public float getDeathlyAlpha(){
    return map(
      constrain((remainingStickAroundTime / stickAroundTime), 0, 1),
      0, 1, 0, 255
    );
  }
  
  public float getEngagementDistance(){
    return radius * 2 + shipRadius;
  }
  
  public void die(float fatalDamage, GameObject damageDealer, int damageType) {
	if ( !aboutToExplode ){
		boolean killedByExplosion = 
			(damageType == EXPLOSIVE_DAMAGE) || ((damageDealer instanceof Bot) && ((Bot)damageDealer).aboutToExplode );
			
		if ( killedByExplosion ) {
			//Chain reaction. Big explosions lead to even bigger ones.
			if ( damageDealer instanceof Bot ){
				explosionComboCount += ((Bot)damageDealer).explosionComboCount;
			}
			
			lightTheFuse(1.0f);
			return; //Delay death until the timed explosion or the next time the ship takes damage
			
		} else if ( damageDealer instanceof Bot ){
			explosionComboCount = 1;
		}
	}
	
	explode();
	
    //Dead ships don't collide with anything
    Shape sh = body.getShapeList();
    while(sh != null){
      sh.getFilterData().maskBits = 0;
      sh = sh.getNext();
    }
    
	//Add points
	awardPoints(points * explosionComboCount, explosionComboCount, getPosition());
	
    //This lets the Director know when to speed up the arrival of the next wave. Note: replace with observer pattern.
    enemiesStillAlive--;
	//Maybe reward the player with a powerup
	maybeSpawnPickup();
	
	super.die(fatalDamage, damageDealer, damageType);
  }
  
  public void lightTheFuse(float countdown){
	aboutToExplode = true;
	explosionCountdown = countdown;
	hitpoints = 1; //Taking any more damage will cause the ship to go boom early.
  }
  
  public void explode(){
	float multiplier = explosionComboCount;
	super.explode(multiplier);
  }
  
  public void maybeSpawnPickup(){
	/*
	Vec2 powerupOffset = new Vec2(random(-radius*0.8, radius*0.8), random(-radius*0.8, radius*0.8));
	Pickup powerup = new PowerPickup( getPosition().addLocal(powerupOffset) );
	entities.add(powerup);
	//*/
  }
  
  public void limitVector(Vec2 vector, float maxLength){
    if ( vector.length() > maxLength ){
        vector.normalize();
        vector.mulLocal(maxLength);
      }
  }
  
}

class BotContact {
  Bot bot;
  ContactPoint contact;
  
  BotContact(Bot bot, ContactPoint contact){
    this.bot = bot;
    this.contact = contact;
  }
}
class Director extends GameObject {
  protected float gameTime = 0;
  protected float nextWaveTime = 0;
  float waveInterval = 20;
  float waveRating = 70;
  int waveNumber = 0;
  int maxEnemies = 40;
  int maxShipsPerWave = 30;
  
  float linearGrowth = 10;
  float exponentialGrowth = 1.01f;
  
  protected List<ShipTemplate> templates;
  
  Director(){
    super();
	drawOnUI = true;
    
    templates = new ArrayList<ShipTemplate>();
	addTemplate("lancer", new LancerFactory(), 1.5f, 4.0f, 20, 50, 90000, 50, 10.0f, 7);
	addTemplate("kamikaze", new KamikazeFactory(), 1.0f, 2.0f, 7, 10, 70000, 70, 30.0f, 5);
    addTemplate("mothership", new MothershipFactory(), 5.0f, 7.0f, 500, 500, 900000, 30, 50.0f, 2);
  }
  
  public void update(float dt){
	//Pause the waves if there are too many enemies already on screen
	if ( enemiesStillAlive > maxEnemies ){
		return;
	}
	
	if ( enemiesStillAlive > 0 ){
	  gameTime += dt;
	} else {
	  gameTime += dt * 2.5f;
	}

    if (nextWaveTime <= gameTime){
      spawnWave(waveRating);
      nextWaveTime = gameTime + waveInterval;
      waveRating = (waveRating + linearGrowth) * exponentialGrowth;
    }    
  }
  
  public void spawnWave(float rating){
    waveNumber++;
    int spawnedShips = 0;
	
    float availableRating = rating;
    while ( (availableRating > 0) && (spawnedShips < maxShipsPerWave) ){
      ShipTemplate template = pickRandomTemplate(availableRating);
      if ( template == null ){
        break;
      }

      int shipCount = min(floor(availableRating / template.averageRating), template.maxShips);
      Vec2 regionCorner = new Vec2(
        random(template.maxRadius + 1, worldHalfWidth - template.maxRadius - 1) * randomSign(), 
        random(template.maxRadius + 1, worldHalfHeight - template.maxRadius - 1) * randomSign()
      );
      float regionWidth = random(template.maxRadius * sqrt(shipCount) * 2, template.maxRadius * shipCount * 2);
      float regionHeight = random(template.maxRadius * sqrt(shipCount) * 2, template.maxRadius * shipCount * 2);
	  float spawnDelay = random(0, waveInterval/2);
	  float spawnDelayVariance = 1.0f;
	  float spawnDuration = 3;
      
      for(int i = 0; i < shipCount; i++){
        //Randomize ship properties based on the template
        float radius = random(template.minRadius, template.maxRadius);
        float maxHitpoints = map(radius, template.minRadius, template.maxRadius, template.minHitpoints, template.maxHitpoints);
		
        //Place the ship somewhere inside the selected region
        Vec2 position = new Vec2(random(0, regionWidth), random(0, regionHeight));
        position.addLocal(regionCorner);
        position.x = constrain(position.x, -worldHalfWidth + radius, worldHalfWidth - radius);
        position.y = constrain(position.y, -worldHalfHeight + radius, worldHalfHeight - radius);
        
        Ship ship = template.factory.newShip(position, radius, maxHitpoints);
        ship.maxThrust = template.thrust;
        ship.maxSpeed = template.speed;
        
        entities.add( new ShipSpawner(ship, spawnDelay + random(0, spawnDelayVariance), spawnDuration) );
		
		spawnedShips++;
		if ( spawnedShips > maxShipsPerWave ){
			break;
		}
      }
      
      availableRating = availableRating - shipCount * template.averageRating;
    }
  }
  
  public void addTemplate(
    String debugName,
    ShipFactory factory,
    float minRadius, float maxRadius,
    float minHitpoints, float maxHitpoints,
    float thrust,
    float speed,
    float averageRating,
    int maxShips
  ){
    ShipTemplate template = new ShipTemplate(debugName, factory, minRadius, maxRadius, minHitpoints, maxHitpoints, thrust, speed, averageRating, maxShips);
    templates.add(template);
  }
  
  public ShipTemplate pickRandomTemplate(float maxRating){
    List<ShipTemplate> usableTemplates = new ArrayList<ShipTemplate>();
    for(ShipTemplate template : templates){
      if (template.averageRating <= maxRating){
        usableTemplates.add(template);
      }
    }
    if ( usableTemplates.size() > 0 ) {
      return usableTemplates.get( PApplet.parseInt(random(0, usableTemplates.size())) );
    } else {
      return null;
    }    
  }
  
  public void drawUI(){
    pushMatrix();
      rectMode(CORNER);
      int border = 10;
	  
      float waveWait = constrain((nextWaveTime - gameTime)/waveInterval, 0, 1);
      float waveIndicatorWidth = 100, waveIndicatorHeight = 10;
      translate(width - waveIndicatorWidth - border, border);
	  
      rectMode(CORNER);
      stroke(0);
      fill(255);
      rect(0, 20, waveIndicatorWidth, waveIndicatorHeight);
      noStroke();
      fill(50, 50, 50);
      rect(0, 20, waveIndicatorWidth * (waveWait), waveIndicatorHeight);
	  
	  textFont(font);
  	  textAlign(LEFT, TOP);
	  textSize(20);
	  fill(0);
      text("WAVE: " + waveNumber, 0, 0);
      
    popMatrix();
  }
}

class ShipTemplate {
  String debugName;
  ShipFactory factory;
  float minRadius, maxRadius;
  float minHitpoints, maxHitpoints;
  float thrust;
  float speed;
  float averageRating;
  int maxShips;
  
  ShipTemplate(
    String debugName,
    ShipFactory factory,
    float minRadius, float maxRadius,
    float minHitpoints, float maxHitpoints,
    float thrust,
    float speed,
    float averageRating,
    int maxShips
  ){
    this.debugName = debugName;
    this.factory = factory;
    this.minRadius = minRadius;
    this.maxRadius = maxRadius;
    this.minHitpoints = minHitpoints;
    this.maxHitpoints = maxHitpoints;
    this.thrust = thrust;
    this.speed = speed;
    this.averageRating = averageRating;
    this.maxShips = maxShips;
  }
}

public float randomSign(){
  if ( random(0.0f, 1.0f) > 0.5f ){
    return 1;
  } else {
    return -1;
  }
}

class ShipSpawner extends GameObject {
  Ship shipToSpawn;
  float spawnDelay, spawnDuration, elapsedTime = 0;
  
  ShipSpawner(Ship shipToSpawn, float spawnDelay, float spawnDuration){
    this.shipToSpawn = shipToSpawn;
    this.spawnDelay = spawnDelay;
    this.spawnDuration = spawnDuration;
  }
  
  public void update(float dt){
    if ( dead ){
      return;
    }
    
    elapsedTime += dt;
    if ( elapsedTime >= spawnDelay + spawnDuration ){
      spawnTheShip();
      die();
    }
  }
  
  public void draw(){
    if ( elapsedTime < spawnDelay ){
      return;
    }
    pushMatrix();
      translate(shipToSpawn.startPosition.x, shipToSpawn.startPosition.y);
      float phase = constrain((elapsedTime - spawnDelay) / spawnDuration, 0.0f, 1.0f);
      fill( shipToSpawn.baseColor, map(phase, 0, 1, 80, 250));
      noStroke();
      ellipse(0, 0, shipToSpawn.radius*2*phase*1.05f, shipToSpawn.radius*2*phase*1.05f);
    popMatrix();
  }
  
  public void spawnTheShip(){
    entities.add(shipToSpawn);
  }
}

class ShipFactory{
  public Ship newShip(Vec2 position, float radius, float hitpoints){
    return null;
  }
}

class MothershipFactory extends ShipFactory{
  public Mothership newShip(Vec2 position, float radius, float hitpoints){
    return new Mothership(position, radius, hitpoints);
  }
}

class LancerFactory extends ShipFactory{
  public Lancer newShip(Vec2 position, float radius, float hitpoints){
    return new Lancer(position, radius, hitpoints);
  }
}

class KamikazeFactory extends ShipFactory{
  public Kamikaze newShip(Vec2 position, float radius, float hitpoints){
    return new Kamikaze(position, radius, hitpoints);
  }
}
class Explosion extends GameObject {
  Vec2 position;
  float radius;
  float impulse;
  float damage;
  float impulseFalloff;
  float damageFalloff;
  float lifetime;
  float remainingLifetime;
  int fillColor = color(255, 100, 100);
  
  GameObject creator = null; //What/who caused this explosion. Can be null.
  
  protected boolean damageDone = false;
  
  Explosion(){
    radius = 1;
    impulse = 4000;
    damage = 5;
    impulseFalloff = 1;
    damageFalloff = 1;
    remainingLifetime = lifetime = 5;
  }
  
  public void setLifetime(float lifetime){
    this.remainingLifetime = this.lifetime = lifetime;
  }
  
  public void update(float dt){
    if ( lifetime > 0 ){
      remainingLifetime -= dt;
    }
    
    if ( remainingLifetime <= 0 ){
      die();
    }
    
	if ( !damageDone ){
		WorldQuery targets = queryEntitiesInRange();
		for(PhysicalObject target : targets){
		  if (target.dead || (target.body == null) || !(target instanceof Ship) ){
			continue;
		  }
		  
		  Ship targetShip = (Ship)target;
		  
		  Vec2 direction = targetShip.getPosition().sub(position);
		  float distance = direction.length();
		  
		  //Check whether the target is *really* in range
		  if ( distance > radius + targetShip.radius ){
			continue;
		  }
		  		  
		  direction.normalize();
		  targetShip.body.applyImpulse(direction.mul(impulseAtDistance(distance)), targetShip.getPosition());
		  
		  float scaledDamage = damageAtDistance(distance);
		  targetShip.takeDamage(scaledDamage, creator, EXPLOSIVE_DAMAGE);
		}
		damageDone = true;
	}
  }
  
  public void draw(){
    pushMatrix();
    translate(position.x, position.y);
    float a = map(max(remainingLifetime / lifetime, 0), 0, 1, 5, 200);
    fill(fillColor, a);
    noStroke();
    ellipse(0, 0, radius * 2, radius * 2);
    popMatrix();
  }
  
  public WorldQuery queryEntitiesInRange(){
    AABB bounds = new AABB(new Vec2(position.x - radius, position.y - radius), new Vec2(position.x + radius, position.y + radius));
    return new AABBQuery(bounds, world, false, 50);
  }
  
  public float damageAtDistance(float distance){
    return damage * ( distance / radius * damageFalloff );
  }
  
  public float impulseAtDistance(float distance){
    return impulse;
  }
}
class GameObject {
  boolean dead = false;
  float timeOfDeath = 0; //timestamp, seconds
  float stickAroundTime = 0; //How long to stick around after death
  float remainingStickAroundTime = 0;
  
  boolean drawOnUI = false; //Whether to call the object's drawUI() method when rendering the UI. Only set this flag once!
                            //The director doesn't re-check it after the entity has been added.
  
  public void handleInput(){
  }
  
  public void update(float dt){
  }
  
  public void draw(){
  }
  
  public void drawUI(){
  }
  
  public void deInit(){
    //stub
  }
  
  public void die(){
    dead = true;
    timeOfDeath = millis() / 1000;
    remainingStickAroundTime = stickAroundTime;
  }
}

class EntityBag {
  private List<GameObject> entities, entitiesWithUI, entitiesToAdd, entitiesToRemove;
 
  EntityBag(){
    entities = new ArrayList<GameObject>();
	entitiesWithUI = new ArrayList<GameObject>();
    entitiesToAdd = new ArrayList<GameObject>();
    entitiesToRemove = new ArrayList<GameObject>();
  }
  
  public void add(GameObject entity){
    entitiesToAdd.add(entity);
	if ( entity.drawOnUI ){
		entitiesWithUI.add(entity);
	}
  }
  
  public void remove(GameObject entity){
    entitiesToRemove.add(entity);
  }
  
  public List<GameObject> getAll(){
    return entities;
  }
  
  public void update(float dt){
    float currentTime = millis() / 1000.0f;
    
    for(GameObject entity : entities){
      entity.update(dt);
      if ( entity.dead ){
        entity.remainingStickAroundTime -= dt;
        if ( entity.remainingStickAroundTime <= 0 ){
          entitiesToRemove.add(entity);
        }
      }
    }
    
    //Warning: Inefficient! Consider optimization later.
    for(GameObject entity : entitiesToRemove){
	  entity.deInit();
      entities.remove(entity);
	  if ( entity.drawOnUI ){
		entitiesWithUI.remove(entity);
	  }
    }
    entitiesToRemove.clear();
    
    for(GameObject entity : entitiesToAdd){
	  if ( entity instanceof PhysicalObject ){
		((PhysicalObject)entity).addTo(world);
	  }
      entities.add(entity);
    }
    entitiesToAdd.clear();
  }
  
  public void draw(){
    for(GameObject entity : entities){
      entity.draw();
    }
  }
  
  public void drawUI(){
    for(GameObject entity : entitiesWithUI){
      entity.drawUI();
    }
  }
  
  public void handleInput(){
    for(GameObject entity : entities){
      entity.handleInput();
    }
  }
}
class LinkedProgressBar extends GameObject {
	GameObject subject;
	float barWidth, barHeight;
	PVector position;
	int borderColor, backgroundColor, valueColor;
	boolean showLabel = false;
	
	protected float currentValue = 0, targetValue, maxValue = 100;
		
	LinkedProgressBar(GameObject subject, float barWidth, float barHeight, PVector position){
		drawOnUI = true;
	
		this.subject = subject;
		this.barWidth = barWidth;
		this.barHeight = barHeight;
		this.position = position;
		
		borderColor = color(0);
		backgroundColor = color(255);
		valueColor = color(255, 20, 20);
		
		readSubjectValues();
	}
	
	public void readSubjectValues(){
		//read the current value of the to-be-displayed property here
	}
	
	public void update(float dt){
		readSubjectValues();
		currentValue = targetValue; //Could also use some sort of easing algorithm for prettier animation
	}
	
	public void drawUI(){
		//Display the progress bar
		if ( subject != null ){
			pushMatrix();
			  translate(position.x, position.y);
			  drawBar();
			  if ( showLabel ){
				drawLabel();
			  }
			popMatrix();
		}
	}
	
	public void drawBar(){
		strokeWeight(1);
		stroke(borderColor);
		fill(backgroundColor);
		rectMode(CORNER);
		rect(0, 0, barWidth, barHeight);

		noStroke();
		fill(getFillColor());
		rect(0, 0, max(barWidth * getPercentage(), 0), barHeight);
	}
	
	public void drawLabel(){
		fill(0);
		textAlign(CORNER, CENTER);
		textFont(font);
		textSize(20);
		text(ceil(currentValue), getLabelX(), barHeight/2);
	}
	
	public float getPercentage(){
		return constrain((currentValue / maxValue), 0, 1);
	}
	
	public float getLabelX(){
		return barWidth + 10;
	}
	
	public int getFillColor(){
		return valueColor;
	}
}

class HealthBar extends LinkedProgressBar {
	HealthBar(Ship subject, float barWidth, float barHeight, PVector position){
		super(subject, barWidth, barHeight, position);
	}

	HealthBar(Ship subject){
		this(subject, 200, 12, new PVector(10, height - 22));
	}
	
	public void readSubjectValues(){
		targetValue = ((Ship)subject).hitpoints;
		maxValue = ((Ship)subject).maxHitpoints;
	}
}

class PowerBar extends LinkedProgressBar {
	PowerBar(PlayerShip subject, float barWidth, float barHeight, PVector position){
		super(subject, barWidth, barHeight, position);
		valueColor = color(50, 50, 255);
	}

	PowerBar(PlayerShip subject){
		this(subject, 200, 12, new PVector(width - 210, height - 22));
	}
	
	public void readSubjectValues(){
		targetValue = ((PlayerShip)subject).power;
		maxValue = ((PlayerShip)subject).maxPower;
	}
	
	public int getFillColor(){
		if (targetValue < repulsorCost){
			return lerpColor(color(100), color(150, 150, 150), targetValue / repulsorCost);
		} else {
			return valueColor;
		}
	}
	
	public float getLabelX(){
		return -30;
	}
	
	public void drawBar(){
		super.drawBar();
		float tickPosition = repulsorCost;
		stroke(0);
		while ( tickPosition < maxValue ){
			final float displayPosition = barWidth * (tickPosition / maxValue);
			line( displayPosition, 0, displayPosition, barHeight );
			tickPosition += repulsorCost;
		}
	}
}
public void handleInput(){
	mouse = MouseInfo.getPointerInfo().getLocation(); 

	mouseDx = mouse.x - prevMouse.x;
	mouseDy = mouse.y - prevMouse.y;
	prevMouse = mouse;

	final int mouseBox = 100;
	final int mouseMinX = width/2 - mouseBox;
	final int mouseMaxX = width/2 + mouseBox;
	final int mouseMinY = height/2 - mouseBox;
	final int mouseMaxY = height/2 + mouseBox;

	//Ensure the mouse doesn't leave the screen
	if ( (mouseX < mouseMinX) || (mouseX > mouseMaxX) || (mouseY < mouseMinY) || (mouseY > mouseMaxY) ){
		prevMouse.x = frame.getLocation().x + width/2;
		prevMouse.y = frame.getLocation().y + height/2;
		robot.mouseMove(prevMouse.x, prevMouse.y);
	}
	
	entities.handleInput();
}

public void mouseWheel(int wheelDelta){
  if ( playerShip == null ){
	return;
  }

  //Up = lengthen, down = shorten
  wheelDelta = - wheelDelta;
  float delta = 0;
  
  if ( wheelDelta > 0 ){
    delta = beamExtensionSpeed * wheelDelta;
  } else {
    delta = beamContractionSpeed * wheelDelta;
  }
  
  playerShip.beamJoint.m_length = constrain(
    playerShip.beamJoint.m_length + delta, 
    minBeamLength + shipRadius + hookRadius, 
    maxBeamLength + shipRadius + hookRadius
  );
}
class Kamikaze extends Bot {
  
  Kamikaze(Vec2 startPosition, float radius, float maxHitpoints){
    super(startPosition, radius, maxHitpoints);
    
    maxThrust = 70000; //70 000
    maxSpeed = 70;
    
    baseColor = color(255, 50, 50);
	points = 20;
	
	explosionDamage = 50;
	explosionRadiusFactor = 3;
	explosionImpulse = 5000;
  }  
  
  Kamikaze(Vec2 startPosition, float radius){
    this(startPosition, radius, 10);
  }
  
  public void onCollisionResult(ContactResult result, PhysicalObject otherObject){
    //Self-destruct when we hit the player's ship
    if ( !dead && (otherObject instanceof PlayerShip) ){
	  PlayerShip player = (PlayerShip)otherObject;
	  player.takeDamage(explosionDamage, this, KAMIKAZE_DAMAGE);
	  aboutToExplode = true;
	  die(); 
    } else {
      super.onCollisionResult(result, otherObject);
    }
  }
  
  public void drawHull(){
	if ( !dead ){
		stroke(150, 0, 0);
		fill(255, 50, 50);
    } else {
        stroke(180, getDeathlyAlpha());
        noFill();
    }
      
    rectMode(CORNER);
	float rectSize = radius * 0.9f;
	rect(-rectSize, -rectSize, rectSize*2, rectSize*2);
	rotate(radians(45));
	rect(-rectSize, -rectSize, rectSize*2, rectSize*2);
	  
    ellipse(0, 0, radius*2, radius*2);
  }
}
class Lancer extends Bot {
  Lancer(Vec2 startPosition, float radius, float maxHitpoints){
    super(startPosition, radius, maxHitpoints);
    baseColor = color(100, 100, 255);
	points = 10;
  }
  
  public void drawHull(){
	final float relativeDmg = 1 - hitpoints/maxHitpoints;
  
    if ( !dead ){
      stroke( lerpColor( color(10, 10, 170), color(170, 10, 10), relativeDmg ) );
      fill(255);
    } else {
      stroke(180, getDeathlyAlpha());
      noFill();
    }
    ellipse(0, 0, radius*2, radius*2);

    rotate(radians(45));
    
    if ( !dead ){
	  int clr = lerpColor(baseColor, color(255, 50, 50), relativeDmg);
      stroke(clr);
      fill(clr);
    }
    rectMode(CENTER);
    rect(0, 0, radius/1.05f, radius/1.1f);
  }
  
}
class Mothership extends Bot {
  Vec2 desiredPosition;
  float spawnPeriod = 10;
  float remainingSpawnCooldown = 5;
  
  Vec2 nextKamikazePosition;
  float nextKamikazeRadius;
  float minSpawnGap = 1.0f;
  
  Mothership(){
    super();
  }
  
  Mothership(Vec2 startPosition, float radius, float maxHitpoints){
    super(startPosition, radius, maxHitpoints);
    desiredPosition = this.startPosition;
    density = enemyShipDensity * 3;
    maxThrust = 900000;
    maxSpeed = 30;
    
    baseColor = color(255, 179, 0);
	points = 100;
    
    spawnPeriod = random(8, 10);
    remainingSpawnCooldown = spawnPeriod + random(0, 5);
    
    //Pick the parameters for the first kamikaze to be spawned
    nextKamikazeRadius = random(1.0f, 2.0f);
    nextKamikazePosition = pickKamikazePosition(minSpawnGap, nextKamikazeRadius);
  }  
  
  Mothership(Vec2 startPosition, float radius){
    this(startPosition, radius, 500);
  }
  
  public void update(float dt){
    if ( !(dead || isTractored) ){
		//Spawn kamikazes every X seconds
		remainingSpawnCooldown -= dt;
		if ( remainingSpawnCooldown <= 0 ){
		  nextKamikazePosition.addLocal(getPosition());
		  spawnKamikaze(nextKamikazePosition, nextKamikazeRadius);
		  
		  //Choose the parameters for the next one
		  nextKamikazeRadius = random(1.0f, 2.0f);
		  nextKamikazePosition.set(pickKamikazePosition(minSpawnGap, nextKamikazeRadius));
		  
		  remainingSpawnCooldown = spawnPeriod + remainingSpawnCooldown;
		}
    }
	
    super.update(dt);
  }
  
  public Vec2 calculateSteeringVector(){
	//Stay close to the original position
	Vec2 steering = arrive(desiredPosition, 20);
	//Avoid collisions with other Motherships
	steering.addLocal( separate(3, entities.getAll(), Mothership.class).mul(2) );	
	
	return steering;
  }
  
  public void draw(){
    if ( !dead ){
      drawNextKamikaze();
    }
    super.draw();
  }
  
  public void drawHull(){
    if ( !dead ){
      stroke(0);
      fill(baseColor);
    } else {
      stroke(180, getDeathlyAlpha());
      noFill();
    }
    
    ellipse(0, 0, radius*2, radius*2);
    rectMode(CENTER);
    rect(0, -radius/2, 1.0f, radius);
    fill(255, 50, 50);
    
    //Display hitpoints as a growing bar
    if ( !dead ){
      noStroke();
      rect(0, -radius/2, 0.8f, map(hitpoints, 0, maxHitpoints, radius, 0) );
    }
  }
  
  public void drawNextKamikaze(){
    pushMatrix();
      Vec2 position = body.getPosition();
      translate(position.x, position.y);
      
      float phase = constrain(remainingSpawnCooldown / spawnPeriod, 0, 1);
      stroke(255, 50, 50, map(phase, 0, 1, 255, 0));
      fill(255, 50, 50, map(phase, 0, 1, 255, 0));
      
      line(0, 0, nextKamikazePosition.x, nextKamikazePosition.y);

      float growingDiameter = nextKamikazeRadius * (1 - phase) * 2;
      ellipse(nextKamikazePosition.x, nextKamikazePosition.y, growingDiameter, growingDiameter);
    popMatrix();
  }
  
  public void spawnKamikaze(Vec2 kPosition, float kRadius){
    Kamikaze bot = new Kamikaze(
      kPosition,
      kRadius
    );
    entities.add(bot);
  }
  
  public Vec2 pickKamikazePosition(float minGap, float kamikazeRadius){
    //Pick a random spot somewhere just outside the mothership's radius
    Vec2 kamikazePosition = new Vec2(0, radius + minGap + kamikazeRadius);
    kamikazePosition = rotateVector(kamikazePosition, radians(random(0, 360)));
    return kamikazePosition;
  }
  
  public void maybeSpawnPickup(){
	Vec2 powerupOffset = new Vec2(random(-radius*0.8f, radius*0.8f), random(-radius*0.8f, radius*0.8f));
	Pickup powerup = new HealthPickup( getPosition().addLocal(powerupOffset) );
	entities.add(powerup);
  }
}
class Particle {
	Vec2 position, velocity, acceleration;
	float damping = 0.0f;
	float age, lifetime;
	int startColor = color(100, 100, 100, 200);
	int endColor = color(100, 100, 100, 0);
	float size = 1;
	boolean dead = false;
	
	Particle(){
		//...stub
	}
	
	public void update(float dt){
		position.addLocal(velocity.mul(dt));
		velocity.addLocal(acceleration.mul(dt));
		Vec2 dampingAcc = velocity.mul(-damping * dt);
		velocity.addLocal(dampingAcc);
		
		age += dt;
		if ( age > lifetime ){
			dead = true;
		}
	}
	
	public void draw(){
		fill(getColor());
		noStroke();
		ellipse(position.x, position.y, size, size);
	}
	
	public int getColor(){
		return lerpColor(startColor, endColor, constrain(age/lifetime, 0, 1));
	}
}

class ParticleFactory {
	public Particle newParticle(){
		return new Particle();
	}
}

class CollisionParticle extends Particle {
	CollisionParticle(){
		startColor = color(255, 110, 2, 250);
		endColor = color(255, 255, 0, 0);
	}
	
	public void draw(){
		drawStraightTrail();
		super.draw();
	}
	
	public void drawStraightTrail(){
		float trailTime = min(0.45f, age);
		Vec2 trailStart = position.sub(velocity.mul(trailTime));
		stroke(getColor());
		line(position.x, position.y, trailStart.x, trailStart.y);
	}
}

class CollisionParticleFactory extends ParticleFactory {
	public CollisionParticle newParticle(){
		return new CollisionParticle();
	}
}

class TextParticle extends Particle {
	String content = "[NO TEXT]";
	
	TextParticle(){
		super();
		startColor = color(0, 250);
		endColor = color(0, 1);
		size = 24;
	}
	
	public void draw(){
		pushMatrix();
			translate(position.x, position.y);
			scale(1.0f / worldToScreenScale);
			textFont(scoreFont);
			fill(getColor());
			textSize(size);
			text(content, 0, 0);
		popMatrix();
	}
}

class TextParticleFactory extends ParticleFactory {
	public TextParticle newParticle(){
		return new TextParticle();
	}
}

class ParticleEmitter extends GameObject {
	protected ArrayList<Particle> particles;
	int limit = 500;
	
	/* Particle properties */
	FloatRange pPosition;      //How much particle position varies from emitter position
	VectorRange pDirection, pAcceleration;
	FloatRange pVelocity, pLifetime, pDamping;
	ParticleFactory factory;
	
	/* Emitter properties */
	Vec2 position;
	float rate;                //particles per second
	float emitterLifetime = 0; //0 = forever
	float emitterAge = 0;	
	PhysicalObject leader = null; //follow this object around
	float elapsedTime = 0;
	
	ParticleEmitter(ParticleFactory factory){
		this.factory = factory;
		particles = new ArrayList<Particle>();
		position = new Vec2(0, 0);
		pPosition = new FloatRange(-1, 1);
		pDirection = new VectorRange(new Vec2(0, -1), radians(360));
		pVelocity = new FloatRange(5.0f, 10.0f);
		pAcceleration = new VectorRange(new Vec2(0, 0), 0);
		pLifetime = new FloatRange(3, 5);
		pDamping = new FloatRange(0.1f, 0.5f);
		rate = 20;
	}	
	
	public void update(float dt){
		followTheLeader();
		if ( dead ){
			return;
		}
		
		spawnParticles(dt);
		updateParticles(dt);
	}
	
	public void followTheLeader(){
		if ( !dead && (leader != null) && (leader.body != null) ){
			if ( leader.dead ){
				die();
				return;
			} else {
				position = leader.getPosition();
			}
		}
	}
	
	public void spawnParticles(float dt){
		//Spawn new particles
		elapsedTime += dt;
		float particleInterval = 1.0f / rate;
		while( elapsedTime > particleInterval ){
			emitParticle();
			elapsedTime -= particleInterval;
		}
	}
	
	public void updateParticles(float dt){
		//Update particles
		for(int i = particles.size()-1; i >= 0; i--){
			Particle particle = particles.get(i);
			particle.update(dt);
			if ( particle.dead ){
				particles.remove(i);
			}
		}
	}
	
	public void draw(){
		for(Particle particle : particles){
			particle.draw();
		}
	}
	
	public Particle emitParticle(){
		Particle particle = makeParticle();
		
		if ( particles.size() < limit ){
			particles.add(particle);
		}
		
		return particle;
	}
	
	public Particle makeParticle(){
		Particle particle = factory.newParticle();
		
		particle.position = position.add(new Vec2(pPosition.getRandom(), pPosition.getRandom()));
		particle.velocity = pDirection.getRandom().mul(pVelocity.getRandom());
		particle.acceleration = pAcceleration.getRandom();
		particle.lifetime = pLifetime.getRandom();
		particle.damping = pDamping.getRandom();
		
		return particle;
	}
}

class TrailEmitter extends ParticleEmitter {
	Vec2 lastParticlePosition = null;
	
	TrailEmitter(TrailParticleFactory factory){
		super(factory);
		pPosition = new FloatRange(0, 0);
		pDirection = new VectorRange(new Vec2(0, 0), 0);
		pVelocity = new FloatRange(0, 0);
		pAcceleration = new VectorRange(new Vec2(0, 0), 0);
		pLifetime = new FloatRange(3, 5);
		pDamping = new FloatRange(0, 0);
		rate = 20;
	}
	
	public void spawnParticles(float dt){
		if ( (leader == null) || (leader.body == null) ){
			return;
		}
		
		if ( lastParticlePosition == null ){
			lastParticlePosition = position.clone();
			return;
		}
		
		Ship ship = (Ship)leader;
		
		TrailParticle particle = (TrailParticle)makeParticle();
		particle.previousPosition = lastParticlePosition;
		particle.position = position;
		particle.size = ship.radius;
		particle.lifetime = 0.2f;
		particle.startColor = ship.baseColor;
		particle.endColor = color(255, 0);
		particles.add(particle);
		
		lastParticlePosition = position;
	}
}

class TrailParticle extends Particle {
	Vec2 previousPosition;

	TrailParticle(){
		super();
	}
	
	public void draw(){
		fill(getColor());
		noStroke();
		quad(
			position.x, position.y - size,
			position.x, position.y + size,
			previousPosition.x, previousPosition.y + size,
			previousPosition.x, previousPosition.y - size
		);
	}
}

class TrailParticleFactory extends ParticleFactory {
	public TrailParticle newParticle(){
		return new TrailParticle();
	}
}

class FloatRange {
	float high, low;
	
	FloatRange(float high, float low){
		this.high = max(low, high);
		this.low = min(low, high);
	}
	
	public float getRandom(){
		return random(low, high);
	}
}

class VectorRange {
	Vec2 vector;
	float spread; //in radians
	protected float halfSpread;
	
	VectorRange(Vec2 vector, float spread){
		this.vector = vector;
		this.spread = spread;
		halfSpread = spread/2;
	}
	
	public Vec2 getRandom(){
		return rotateVector(vector, random(-halfSpread, halfSpread));
	}
}

class PositionRange {
	Vec2 vector;
	float spread; //radius / half-square
	protected float halfSpread;
	
	PositionRange(Vec2 vector, float spread){
		this.vector = vector;
		this.spread = spread;
		halfSpread = spread/2;
	}
	
	public Vec2 getRandom(){
		Vec2 result = vector.clone();
		result.x += random(-halfSpread, halfSpread);
		result.y += random(-halfSpread, halfSpread);
		return result;
	}
}

class PhysicalObject extends GameObject {
  Body body;
  
  protected Vec2 startPosition; //Where to place the object when it's first added to the Box2D world
  
  public void addTo(World world){
    //create the body and add it to the world here
  }
  
  public Vec2 getPosition(){
    if ( body != null ){
      return body.getPosition();
    } else {
      return new Vec2(0, 0);
    }
  }
  
  public void deInit(){
    if ( body != null ){
      world.destroyBody(body);
      body = null;
    }
  }
  
  public void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
    //stub
  }  
  
  public void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
    //stub
  }
  
  public void onCollisionResult(ContactResult result, PhysicalObject otherObject){
    //stub
  }
}

public void initPhysics(){
  //Create the Box2d world
  AABB worldBounds = new AABB(new Vec2(-100f, -100f), new Vec2(100f, 100f));
  boolean doSleep = true;
  world = new World(worldBounds, gravity, doSleep);
  
  //Add walls to the world
  walls = new Body[4];
  
  BodyDef wallBodyDef = new BodyDef();
  PolygonDef wallBox = new PolygonDef();
  wallBox.filter.groupIndex = -1; //walls don't collide with each other
  
  wallBox.setAsBox(wallThickness, height/worldToScreenScale * 0.5f);
  wallBodyDef.position.set(-width/worldToScreenScale*0.5f - wallThickness, 0f);
  walls[0] = world.createBody(wallBodyDef);
  walls[0].createShape(wallBox);
  
  wallBodyDef.position.set(width/worldToScreenScale*0.5f + wallThickness, 0f);
  walls[1] = world.createBody(wallBodyDef);
  walls[1].createShape(wallBox);
  
  wallBox.setAsBox(width/worldToScreenScale * 0.5f, wallThickness);
  wallBodyDef.position.set(0f, -height/worldToScreenScale*0.5f - wallThickness);
  walls[2] = world.createBody(wallBodyDef);
  walls[2].createShape(wallBox);
  
  wallBodyDef.position.set(0f, height/worldToScreenScale*0.5f + wallThickness);
  walls[3] = world.createBody(wallBodyDef);
  walls[3].createShape(wallBox);

  //Our contact handler/proxy
  contactHandler = new MyContactListener();
  world.setContactListener(contactHandler);
}

public void updatePhysics(float dt){
  world.step(dt, 8);
}

class MyContactListener implements ContactListener {
  
  private final int COLLISION_ADD = 1;
  private final int COLLISION_PERSIST = 2;
  private final int COLLISION_REMOVE = 3;
  
  public void add(ContactPoint point){
    notifyGameObjects(point, COLLISION_ADD);
    processContact(point);
  }
  
  public void persist(ContactPoint point){
    notifyGameObjects(point, COLLISION_PERSIST);
    processContact(point);
  }
  
  public void remove(ContactPoint point){
    notifyGameObjects(point, COLLISION_REMOVE);
  }
 
  public void result(ContactResult result){
    PhysicalObject obj1 = getObjectFromShape(result.shape1);
    PhysicalObject obj2 = getObjectFromShape(result.shape2);
    
    if ( obj1 != null ){
      obj1.onCollisionResult(result, obj2);
    }
    
    if ( obj2 != null ){
      obj2.onCollisionResult(result, obj1);
    }
	
	spawnCollisionParticles(result);
  }
  
  public void notifyGameObjects(ContactPoint point, int phase){
    PhysicalObject obj1 = getObjectFromShape(point.shape1);
    PhysicalObject obj2 = getObjectFromShape(point.shape2);
     
    switch(phase){
      case COLLISION_ADD:
        if ( obj1 != null ){
          obj1.onCollisionStart(point, obj2);
        }
        
        if ( obj2 != null ){
          obj2.onCollisionStart(point, obj1);
        }
        break;
        
      case COLLISION_PERSIST:
        if ( obj1 != null ){
          obj1.onCollisionPersist(point, obj2);
        }
        
        if ( obj2 != null ){
          obj2.onCollisionPersist(point, obj1);
        }
        break;
        
      case COLLISION_REMOVE:
        break;
    }
  }
  
  public void spawnCollisionParticles(ContactResult result){
	float impactStrength = calculateImpactStrength(result);
	
	int numParticles = ceil(impactStrength / 10); 
	if ( numParticles < 2 ){
		return;
	}
	
	numParticles = ceil(constrain(numParticles, 12, 64));
	
	float averageSpeed = map(impactStrength, 0, 100, 10, 30);
	final float speedVariance = 0.30f;
	FloatRange speed = new FloatRange(averageSpeed * (1 - speedVariance), averageSpeed * (1 + speedVariance));
	
	float averageLifetime = constrain(10 / averageSpeed, 0.2f, 0.8f);
	final float lifetimeVariance = 0.30f;
	FloatRange lifetime = new FloatRange(averageLifetime * (1 - lifetimeVariance), averageLifetime * (1 + lifetimeVariance));
	
	for (int i = 0; i < numParticles; i++ ){
		Particle particle = collisionEmitter.emitParticle();
		particle.position.set(result.position);
		particle.velocity.normalize();
		particle.velocity.mulLocal( speed.getRandom() );
		particle.lifetime = lifetime.getRandom();
	}
  }
  
  public void processContact(ContactPoint point){
    Body body1 = point.shape1.getBody();
    Body body2 = point.shape2.getBody();
    
    //Process the hook's collisions with enemy ships
    if ( playerShip.hookActive && (playerShip.hookedBot == null) ){
      Body target = null;
      if ( body1 == playerShip.hook ){
        target = body2; 
      } else if ( body2 == playerShip.hook ){
        target = body1;
      }
      
      //Is the target a bot ship?
      if ( (target != null) && (target.getUserData() != null) && (target.getUserData() instanceof Bot) ){
        Bot targetBot = (Bot)target.getUserData();
        //handle this outside the physics step
        if ( !targetBot.dead ){
          possibleTargets.add( new BotContact(targetBot, point) ); 
        }
      }
    }
  }
}

//Get the game object associated with a Box2d body. 
public PhysicalObject getObjectFromBody(Body body){
	if ( body == null ){
	  return null;
	}
	if ( (body.getUserData() != null) && (body.getUserData() instanceof PhysicalObject) ){
	  return (PhysicalObject)body.getUserData();
	}
	return null;    
}

//Get the game object associated with a Box2d shape. If the shape has no associated object, 
//this function will return the game object associated with this shape's body instead.
public PhysicalObject getObjectFromShape(Shape shape){
	if ( shape == null ){
	  return null;
	}
	Object obj = shape.getUserData();
	if ( (obj != null) && (obj instanceof PhysicalObject) ){
	  return (PhysicalObject)obj;
	} else {
	  return getObjectFromBody(shape.getBody());
	}
}

//Rotate a 2D vector around the origin. The angle of rotation should be specified in radians.
//Positive values will produce clockwise rotation, negative - counter-clockwise.
public Vec2 rotateVector(Vec2 vector, float angle){
  float sin_a = sin(angle), cos_a = cos(angle);
  return new Vec2(
    cos_a * vector.x - sin_a * vector.y,
    sin_a * vector.x + cos_a * vector.y
  );
}

//Calculate the approximate "strength" of a collision. Note: the returned value will likely
//have no particular relation to real collision physics. It's intended to be a relative indication
//of how strong one object hit another, not an accurate measure of absorbed energy or some such.
public float calculateImpactStrength(ContactResult result){
	return result.normalImpulse /(result.shape1.getBody().getMass() + result.shape2.getBody().getMass());
}

public abstract class WorldQuery implements Iterator<PhysicalObject>, Iterable<PhysicalObject> {
  private org.jbox2d.dynamics.World world;
  private boolean allowDuplicates = true;
  
  WorldQuery(org.jbox2d.dynamics.World world, boolean allowDuplicates){
    this.world = world;
    this.allowDuplicates = allowDuplicates;
  }
  
  public Iterator<PhysicalObject> iterator(){
    return this;
  }
  
  public boolean hasNext(){
    return false;
  }
  
  public PhysicalObject next(){
    return null;
  }
  
  public void remove(){
    throw new UnsupportedOperationException();
  }
}

/*
 * A helper class for finding all PhysicalObject's that overlap a given AABB
 */
class AABBQuery extends WorldQuery {
  private AABB bounds;
  private List<PhysicalObject> foundEntities;
  private Iterator<PhysicalObject> internalIterator;
    
  AABBQuery(AABB bounds, org.jbox2d.dynamics.World world, boolean allowDuplicates, int maxShapes){
    super(world, allowDuplicates);
    this.bounds = bounds;

    Shape[] foundShapes = world.query(bounds, maxShapes);
    foundEntities = new ArrayList<PhysicalObject>(foundShapes.length);
    for(int i = 0; i < foundShapes.length; i++){
      PhysicalObject data = getObjectFromShape(foundShapes[i]);
      if ( data != null ){
        if ( allowDuplicates || !foundEntities.contains(data) ){
          foundEntities.add((PhysicalObject)data);
        }
      }
    }
    internalIterator = foundEntities.iterator();
  }
  
  AABBQuery(AABB bounds, org.jbox2d.dynamics.World world){
    this(bounds, world, true, 500);
  }
  
  public boolean hasNext(){
    return internalIterator.hasNext();
  }
  
  public PhysicalObject next(){
    return internalIterator.next();
  }
}
class Pickup extends PhysicalObject {
	float radius = 1;
	float age = 0; //How long the pickup has existed.
	float growthTime = 0.5f; //When first spawned, the pickup shall animate to full size within this many seconds
	float lifeTime = 0; //How long the pickup stays around. 0 = no limit.
	float fadeOutTime = 3.0f; //How quickly it fades before death
	int bonusPoints = 5; //Get extra points for picking up a powerup
	
	Pickup(Vec2 position, float radius){
		super();
		this.startPosition = position;
		this.radius = radius;
	}
	
	//Check whether a particular entity is allowed to take this pickup
	public boolean allowPickupBy(GameObject taker){
		return false;
	}

	//Fired when the pickup is picked up by something
	public void onPickup(GameObject taker){
		//implement this in subclasses
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	public void draw(){
		Vec2 position = getPosition();
		fill(50, 200, 50);
		stroke(0, 150, 0);
		ellipse(position.x, position.y, getVisibleRadius()*2, getVisibleRadius()*2);
	}
	
	public float getVisibleRadius(){
		if ( age >= growthTime ){
			return radius;
		} else {
			return age/growthTime*radius;
		}
	}
	
	public int getAlpha(){
		if (lifeTime == 0){
			return 255;
		}
		return round(map( 
			constrain((lifeTime - age)/fadeOutTime, 0, 1),
			0, 1, 0, 255
		));
	}
	
	public void update(float dt){
		age += dt;
		if ( (lifeTime != 0) && (age > lifeTime) ){
			die();
		}
	}
	
	public void addTo(World world){
		//Create the Box2d body corresponding to this pickup and add it to the physical world
		BodyDef pickupBodyDef = new BodyDef();
		pickupBodyDef.position.x = startPosition.x;
		pickupBodyDef.position.y = startPosition.y;
		pickupBodyDef.fixedRotation = true;
		
		CircleDef pickupShape = new CircleDef();
		pickupShape.radius = radius;
		pickupShape.density = 0;
		pickupShape.isSensor = true; //pass through other bodies harmlessly
		
		body = world.createBody(pickupBodyDef);
		body.createShape(pickupShape);
		body.setMassFromShapes();
		body.setUserData(this);
	}
	
	protected void handleCollision(ContactPoint contact, PhysicalObject otherObject){
		if ( dead || (otherObject == null) ){
			return;
		}
		if (allowPickupBy(otherObject)){
			onPickup(otherObject);
			if ( otherObject == playerShip ){
				//Note: bad design. What if something else picks up the powerup?
				awardPoints(bonusPoints, 0, getPosition());
			}
			die();
		}
	}
	
	public void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
		handleCollision(contact, otherObject);
	}  
	  
	public void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
		handleCollision(contact, otherObject);
	}
}

class HealthPickup extends Pickup {
	float healthIncrease = 50;
	
	HealthPickup(Vec2 position, float radius){
		super(position, radius);
		lifeTime = 20;
	}
	
	HealthPickup(Vec2 position){
		this(position, 2.0f);
	}

	//The player ship can  pick up health when it's damaged but not yet dead
	public boolean allowPickupBy(GameObject taker){
		if ( !(taker instanceof PlayerShip) || taker.dead){
			return false;
		};
		PlayerShip player = (PlayerShip)taker;
		return (player.hitpoints < player.maxHitpoints);
	}

	//Heal the player
	public void onPickup(GameObject taker){
		PlayerShip player = (PlayerShip)taker;
		player.heal(healthIncrease);
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	public void draw(){
		Vec2 position = getPosition();
		pushMatrix();		
			translate(position.x, position.y);
			float vradius = getVisibleRadius();
			int valpha = getAlpha();
			//The green base circle
			fill(50, 200, 50, valpha);
			stroke(0, 150, 0, valpha);
			ellipse(0, 0, vradius*2, vradius*2);
			//The cross
			rectMode(CENTER);
			fill(255, valpha);
			noStroke();
			float cw = vradius*1.3f, ch = vradius/2.5f;
			rect(0, 0, cw, ch);
			rect(0, 0, ch, cw);
		popMatrix();
	}
}

class PowerPickup extends Pickup {
	float powerIncrease = 10;
	
	PowerPickup(Vec2 position, float radius){
		super(position, radius);
		lifeTime = 10;
	}
	
	PowerPickup(Vec2 position){
		this(position, 1.0f);
	}

	//Only the player can pick up power
	public boolean allowPickupBy(GameObject taker){
		return (taker instanceof PlayerShip) && (!taker.dead);
	}

	//Give the player more power and +1 HP
	public void onPickup(GameObject taker){
		PlayerShip player = (PlayerShip)taker;
		player.power = constrain(player.power + powerIncrease, 0, player.maxPower);
		player.heal(1);
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	public void draw(){
		Vec2 position = getPosition();
		pushMatrix();		
			translate(position.x, position.y);
			float vradius = getVisibleRadius();
			int valpha = getAlpha();
			//The base circle
			fill(50, 200, 50, valpha);
			stroke(0, 150, 0, valpha);
			ellipse(0, 0, vradius*2, vradius*2);
			//The triangle
			fill(255, valpha);
			noStroke();
			triangle(
				0, -vradius * 0.7f,
				-vradius * 0.6f, vradius * 0.6f,
				vradius * 0.6f, vradius * 0.6f
			);
		popMatrix();
	}
}
class PlayerShip extends Ship {
  Body hook;
  DistanceJoint beamJoint;
  
  boolean hookActive = false;
  Bot hookedBot = null;
  Joint hookJoint = null;
  
  float power = 180;
  float maxPower = 600;
  RepulsorBlast repulsor;

  PlayerShip(){
    super();
    hitpoints = maxHitpoints = playerHealth;
    radius = shipRadius;
    stickAroundTime = 6;
    
	repulsor = new RepulsorBlast(this);
	
	maxSpeed = 100;
  }
  
  public void addTo(World world){
	//The player's ship
    BodyDef shipBodyDef = new BodyDef();
    shipBodyDef.position.set(0.0f, -10.0f);
    shipBodyDef.angularDamping = 0.9f;
    shipBodyDef.linearDamping = 3;
    
    CircleDef shipShape = new CircleDef();
    shipShape.radius = shipRadius;
    shipShape.density = shipDensity;
    shipShape.friction = 0.7f;
    
    body = world.createBody(shipBodyDef);
    body.createShape(shipShape);
    body.setMassFromShapes();
    body.setUserData(this);
    
    //The tractor beam "hook"
    BodyDef hookBodyDef = new BodyDef();
    hookBodyDef.position.x = shipBodyDef.position.x;
    hookBodyDef.position.y = shipBodyDef.position.y - shipRadius - hookRadius - initialBeamLength;
    hookBodyDef.linearDamping = 1;
    hookBodyDef.angularDamping = 0.9f;
    hookBodyDef.fixedRotation = true;
    
    CircleDef hookShape = new CircleDef();
    hookShape.radius = hookRadius;
    hookShape.density = hookDensity;
    hookShape.isSensor = true; //pass through other bodies harmlessly
    
    hook = world.createBody(hookBodyDef);
    hook.createShape(hookShape);
    hook.setMassFromShapes();
    
    //Connect the hook to the ship
    DistanceJointDef def = new DistanceJointDef();
    def.initialize(body, hook, body.getWorldCenter(), hook.getWorldCenter());
    def.frequencyHz = 40.0f;
    beamJoint = (DistanceJoint)world.createJoint(def);
	
	repulsor.addTo(world);
  }
  
  public void draw(){
	//Give the repulsor weapon the chance to draw its GFX
	repulsor.draw();
  
    //Draw the tractor beam
    Vec2 position = playerShip.getPosition();
    Vec2 endPosition = hook.getPosition();
    stroke(dead ? 180 : 100);
    line(position.x, position.y, endPosition.x, endPosition.y);
    stroke(dead ? 180 : 0);
    fill(255);
    
    //Draw the ship
    pushMatrix();
      position = body.getPosition();
      translate(position.x, position.y);
      rotate(body.getAngle());
      stroke(dead ? 180 : 0);
      ellipse(0, 0, radius*2, radius*2);
      rect(0, -radius/2, 1, radius);
    popMatrix();
    
    //Draw the ship's tractor beam hook
    pushMatrix();
      position = hook.getPosition();
      translate(position.x, position.y);
      rotate(hook.getAngle());
      
	  if ( dead ){
		stroke(180);
	  } else if (hookActive) {
        stroke(255, 50, 50);
      } else {
        stroke(0);
      }
      
      ellipse(0, 0, hookRadius*2, hookRadius*2);
      rect(0, -hookRadius/2, 1, hookRadius);
    popMatrix();
  }
  
  public void handleInput(){
	if ( dead ){
		return;
	}
	
	//Steer the ship with the mouse
	Vec2 force = new Vec2(mouseDx*mouseStrength, mouseDy*mouseStrength);
	body.applyForce(force, body.getWorldCenter()); //apply the force at the center of mass

    //Activate the tractor beam when LMB is down
    hookActive = mousePressed && (mouseButton == LEFT);
    
    //Release the tractored object with RMB
    if ( mousePressed && (mouseButton == RIGHT) ){
      if ( hookedBot != null ){
        releaseHookedObject();
      } else {
		/*
        //Debug thing: create an explosion 
        Explosion boom = new Explosion();
        boom.position = hook.getPosition();
        boom.radius = random(1.0, 3.0);
        boom.damage = 0;
        entities.add(boom);
		//*/
      }
    }
	
	//Fire the repulsor with SPACE
	if ( keyPressed && (key == ' ') && repulsor.readyToFire() ){
		repulsor.fire();
	}
  }  
  
  public void update(float dt){
	if ( dead && (remainingStickAroundTime < 1.0f) ){
		goToGameOver();
	}
  
    if ( hookedBot != null && (hookedBot.dead || this.dead)){
      releaseHookedObject();
    }
	
	if ( dead ){
		return;
	}
    
    //Check if the hook should latch onto one of the enemy ships
    if ( !possibleTargets.isEmpty() ){
      BotContact bc = possibleTargets.get(0); //Could be improved by sorting by separation, etc
      possibleTargets.clear();
          
      //Attach the hook to the target
      RevoluteJointDef def = new RevoluteJointDef();
      def.body1 = hook;
      def.localAnchor1 = hook.getLocalCenter();
      def.body2 = bc.bot.body;
      def.localAnchor2 = bc.bot.body.getLocalPoint(bc.contact.position);
      //def.initialize(hook, bc.bot.body, bc.contact.position);
      hookJoint = world.createJoint(def);
      hookedBot = bc.bot;
      bc.bot.isTractored = true;
    }
	
	repulsor.update(dt);
  }
  
  public void releaseHookedObject(){
    world.destroyJoint(hookJoint);
    hookJoint = null;
    hookedBot.isTractored = false;
    hookedBot = null;
  }
  
  public float getCollisionDamage(ContactResult result, PhysicalObject otherObject){
    if ( otherObject == null ){
      return 0;
    }
    
    //Take no damage from tractored ships
    if ( otherObject == hookedBot ){
      return 0;
    }
    
    return super.getCollisionDamage(result, otherObject);
  }
  
  public void takeDamage(float amount, GameObject damageDealer, int damageType){
	//The player's ship is invulnerable to explosive damage
	if ( damageType == EXPLOSIVE_DAMAGE ){
		return;
	} else {
		super.takeDamage(amount, damageDealer, damageType);
	}
  }
  
  public void heal(float amount){
	if ( !dead ){
		hitpoints = constrain(hitpoints + amount, 0, maxHitpoints);
	}
  }

  public void gainPower(float amount){
	power = constrain(power + amount, 0, maxPower);
  }
 
  public void deInit(){
	if ( hookedBot != null ){
		releaseHookedObject();
	}
	super.deInit();
  }
}
public void drawWorld(){
  pushMatrix();
	  translate(width/2, height/2);
	  scale(worldToScreenScale);
	  
	  strokeWeight(2f/worldToScreenScale);
	  rectMode(CENTER);
	  fill(255);
	  stroke(0);
	  
	  //Draw the entities
	  entities.draw();
  popMatrix();
}

public void drawUI(){
  entities.drawUI();
  
  //Display the score
  pushMatrix();
	translate(width / 2, 10);
	textFont(font);
	textAlign(CENTER, TOP);
	textSize(24);
	fill(0); noStroke();
	text("SCORE: " + ceil(score), 0, 0);
  popMatrix();
}
final static int UNKNOWN_DAMAGE = 0;
final static int EXPLOSIVE_DAMAGE = 1;
final static int COLLISION_DAMAGE = 2;
final static int KAMIKAZE_DAMAGE = 3;

class Ship extends PhysicalObject {
  float radius;
  float maxThrust = 10000;
  float maxSpeed = 100;
  boolean isTractored = false;

  float hitpoints = 0;
  float maxHitpoints = 100;
  
  float explosionDamage = 40;
  float explosionRadiusFactor = 1.7f;
  float explosionImpulse = 5000;
  
  int baseColor;
  
  Ship(){
    //stub
    baseColor = color(255, 255, 0);
  }
  
  public void takeDamage(float amount, GameObject damageDealer, int damageType){
    if ( dead ) return;
    
    hitpoints = hitpoints - amount;
    if ( hitpoints <= 0 ){
      die(amount, damageDealer, damageType);
    }
  }
  
  public void takeDamage(float amount){
    takeDamage(amount, null, UNKNOWN_DAMAGE);
  }
  
  public void die(float fatalDamage, GameObject damageDealer, int damageType){
	hitpoints = 0;
	super.die();
  }
  
  public void die(){
    die(0, null, UNKNOWN_DAMAGE);
  }
  
  public void explode(float multiplier){
	Explosion boom = new Explosion();
    boom.radius = constrain(radius * explosionRadiusFactor * multiplier, 1, 30);
    boom.damage = constrain(explosionDamage * sqrt(multiplier), 1, 500);
    boom.impulse = explosionImpulse;
    boom.position = getPosition();
    boom.setLifetime(0.5f);
	boom.creator = this;
    entities.add(boom);
  }
  
  public void onCollisionResult(ContactResult result, PhysicalObject otherObject){
    if (dead){
      return;
    }
    
    //Ignore collisions with other dead things
    if ( otherObject != null && otherObject.dead ){
      return;
    }
    
    //Maybe take damage from the collision
    float damage = getCollisionDamage(result, otherObject);
          
    if ( damage > 1 ){
      takeDamage(damage, otherObject, COLLISION_DAMAGE);
    }
  }
  
  public float getCollisionDamage(ContactResult result, PhysicalObject otherObject){
    if ( body == null ){
      return 0;
    }
    
    //Take no damage from the player's ship if currently tractored by it
    if ( isTractored && (otherObject instanceof PlayerShip) ){
      return 0;
    }
    
    float damage = result.normalImpulse / 8000 * (1 - getArmor());
    //Take less collision damage from scenery, i.e. walls.
    if ( otherObject == null ){
      damage = damage * 0.5f;
    }
    
    return damage;
    /*
    float energy = abs(collisionEnergy(result));
    float damage = energy / 1000000;
    return damage;
    */
  }
  
  public float getArmor(){
    return 0.0f;
  }
  
}


class Weapon extends GameObject {
  Ship owner;
  boolean firing = false;
  
  Weapon(Ship owner){
    super();
    this.owner = owner;
  }
  
  public void fire(){
  }
  
  public boolean readyToFire(){
	return true;
  }
}

class RepulsorBlast extends PhysicalObject {
	float duration = 5.0f;
	float cooldown = 0.001f;
	float cost = repulsorCost;
	float force = 1800000;
	float effectRadius = 50;
	float distanceAttenuationFactor = 30;
	
	float remainingCooldown = 0;
	float remainingDuration = 0;
	
	boolean firing = false;
	PlayerShip owner;

	RepulsorBlast(PlayerShip owner){
		super();
		this.owner = owner;
	}
	
	public boolean readyToFire(){
		return !firing && (remainingCooldown <= 0) && (((PlayerShip)owner).power >= cost);
	}
	
	public void fire(){
		((PlayerShip)owner).power -= cost;
		firing = true;
		remainingDuration = duration;
	}
	
	public void update(float dt){
		if ( owner.dead ){
			firing = false;
			return;
		}
		if ( firing ){
			remainingDuration -= dt;
			if ( remainingDuration <= 0 ){
				firing = false;
				remainingDuration = 0;
				remainingCooldown = cooldown;
			}
		} else if ( remainingCooldown > 0 ){
			remainingCooldown -= dt;
		}
	}
	
	public void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
		if ( firing ){
			applyEffectTo(otherObject);
		}
		
	}  
	  
	public void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
		if ( firing ){
			applyEffectTo(otherObject);
		}
	}
	
	public void applyEffectTo(PhysicalObject thing){
		if ( (thing == null) || (thing == owner) || (thing == this) || (thing.body == null) ){
			return;
		}
		
		//Skip tractored bots
		if( (thing instanceof Ship) && (((Ship)thing).isTractored) ){
			return;
		}
		
		//Push the target away from the player
		Vec2 forceVector = thing.body.getWorldCenter().sub(owner.body.getWorldCenter());
		float distance = max(forceVector.length(), 0.00001f); //prevent division-by-zero below
		distance = distance / distanceAttenuationFactor;
		//quadratic fall-off
		float appliedForce = force / (distance * distance); 
		forceVector.normalize();
		forceVector.mulLocal(appliedForce);
		
		thing.body.applyForce(forceVector, thing.body.getWorldCenter());
	}
	
	public void draw(){
		if ( !firing ){
			return;
		}
		pushMatrix();
		Vec2 center = owner.getPosition();
		translate(center.x, center.y);
		float valpha = 120;
		float fadeOutTime = 1.0f;
		if ( remainingDuration < fadeOutTime ){
			valpha = valpha * constrain(remainingDuration/fadeOutTime, 0, 1);
		}
		fill(100, 100, 255, valpha);
		noStroke();
		ellipse(0, 0, effectRadius*2, effectRadius*2);
		popMatrix();
	}
	
	public void addTo(World world){
		CircleDef sensorShape = new CircleDef();
		sensorShape.radius = effectRadius;
		sensorShape.density = 0;
		sensorShape.isSensor = true; 
		
		Shape sensor = owner.body.createShape(sensorShape);
		sensor.setUserData(this);
	}
}

class PulseWeapon extends Weapon {
  float damage = 10;
  float damageRadius = 3;
  
  float warmupTime = 0.3f;
  float cooldownTime = 0.8f;
  float warmupTimeRemaining = 0;
  float cooldownTimeRemaining = 0;
  boolean firing = false;
  
  PulseWeapon(Bot owner){
    super(owner);
  }
  
  public void fire(){
    if ( (cooldownTimeRemaining > 0) || firing ){
      return;
    }
    firing = true;
    warmupTimeRemaining = warmupTime;
  }
  
  public void update(float dt){
    float delta = dt;
    if ( firing ){
      warmupTimeRemaining -= delta;
      if ( warmupTimeRemaining <= 0 ){
        warmupTimeRemaining = 0;
        //TODO: Actually open fire
        cooldownTimeRemaining = cooldownTime;
        firing = false;
      }
    } else if ( cooldownTimeRemaining != 0 ) {
      cooldownTimeRemaining -= delta;
      if ( cooldownTimeRemaining < 0 ){
        cooldownTimeRemaining = 0;
      }
    }
  }
  
  public void draw(){
    if ( firing ){
      noFill();
      stroke(255, 50, 50);
      ellipse(0, 0, damageRadius*2, damageRadius*2);
    }
  }
}
  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "SlingPrototype" });
  }
}
