class Weapon extends GameObject {
  Ship owner;
  boolean firing = false;
  
  Weapon(Ship owner){
    super();
    this.owner = owner;
  }
  
  void fire(){
  }
  
  boolean readyToFire(){
	return true;
  }
}

class RepulsorBlast extends PhysicalObject {
	float duration = 5.0;
	float cooldown = 0.001;
	float cost = repulsorCost;
	float force = 1800000;
	float effectRadius = 50;
	float distanceAttenuationFactor = 30;
	
	float remainingCooldown = 0;
	float remainingDuration = 0;
	
	boolean firing = false;
	PlayerShip owner;

	RepulsorBlast(PlayerShip owner){
		super();
		this.owner = owner;
	}
	
	boolean readyToFire(){
		return !firing && (remainingCooldown <= 0) && (((PlayerShip)owner).power >= cost);
	}
	
	void fire(){
		((PlayerShip)owner).power -= cost;
		firing = true;
		remainingDuration = duration;
	}
	
	void update(float dt){
		if ( owner.dead ){
			firing = false;
			return;
		}
		if ( firing ){
			remainingDuration -= dt;
			if ( remainingDuration <= 0 ){
				firing = false;
				remainingDuration = 0;
				remainingCooldown = cooldown;
			}
		} else if ( remainingCooldown > 0 ){
			remainingCooldown -= dt;
		}
	}
	
	void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
		if ( firing ){
			applyEffectTo(otherObject);
		}
		
	}  
	  
	void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
		if ( firing ){
			applyEffectTo(otherObject);
		}
	}
	
	void applyEffectTo(PhysicalObject thing){
		if ( (thing == null) || (thing == owner) || (thing == this) || (thing.body == null) ){
			return;
		}
		
		//Skip tractored bots
		if( (thing instanceof Ship) && (((Ship)thing).isTractored) ){
			return;
		}
		
		//Push the target away from the player
		Vec2 forceVector = thing.body.getWorldCenter().sub(owner.body.getWorldCenter());
		float distance = max(forceVector.length(), 0.00001); //prevent division-by-zero below
		distance = distance / distanceAttenuationFactor;
		//quadratic fall-off
		float appliedForce = force / (distance * distance); 
		forceVector.normalize();
		forceVector.mulLocal(appliedForce);
		
		thing.body.applyForce(forceVector, thing.body.getWorldCenter());
	}
	
	void draw(){
		if ( !firing ){
			return;
		}
		pushMatrix();
		Vec2 center = owner.getPosition();
		translate(center.x, center.y);
		float valpha = 120;
		float fadeOutTime = 1.0;
		if ( remainingDuration < fadeOutTime ){
			valpha = valpha * constrain(remainingDuration/fadeOutTime, 0, 1);
		}
		fill(100, 100, 255, valpha);
		noStroke();
		ellipse(0, 0, effectRadius*2, effectRadius*2);
		popMatrix();
	}
	
	void addTo(World world){
		CircleDef sensorShape = new CircleDef();
		sensorShape.radius = effectRadius;
		sensorShape.density = 0;
		sensorShape.isSensor = true; 
		
		Shape sensor = owner.body.createShape(sensorShape);
		sensor.setUserData(this);
	}
}

class PulseWeapon extends Weapon {
  float damage = 10;
  float damageRadius = 3;
  
  float warmupTime = 0.3;
  float cooldownTime = 0.8;
  float warmupTimeRemaining = 0;
  float cooldownTimeRemaining = 0;
  boolean firing = false;
  
  PulseWeapon(Bot owner){
    super(owner);
  }
  
  void fire(){
    if ( (cooldownTimeRemaining > 0) || firing ){
      return;
    }
    firing = true;
    warmupTimeRemaining = warmupTime;
  }
  
  void update(float dt){
    float delta = dt;
    if ( firing ){
      warmupTimeRemaining -= delta;
      if ( warmupTimeRemaining <= 0 ){
        warmupTimeRemaining = 0;
        //TODO: Actually open fire
        cooldownTimeRemaining = cooldownTime;
        firing = false;
      }
    } else if ( cooldownTimeRemaining != 0 ) {
      cooldownTimeRemaining -= delta;
      if ( cooldownTimeRemaining < 0 ){
        cooldownTimeRemaining = 0;
      }
    }
  }
  
  void draw(){
    if ( firing ){
      noFill();
      stroke(255, 50, 50);
      ellipse(0, 0, damageRadius*2, damageRadius*2);
    }
  }
}
